import re as regex

from elody.policies.helpers import get_item
from elody.policies.permission_handler import (
    get_permissions,
    handle_single_item_request,
)
from flask import Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)
from storage.storagemanager import StorageManager  # pyright: ignore


class MediafileDownloadPolicy(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match(
            "^(/elody/v[0-9]+)?/mediafiles/[^/]+/download$", request.path
        ):
            return policy_context

        item = get_item(StorageManager(), user_context.bag, request.view_args)
        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [GetRequestRules]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(item, user_context, request, permissions)
                if access_verdict is not None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        break

            if policy_context.access_verdict:
                return policy_context

        return policy_context


class GetRequestRules:
    def apply(
        self, item, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "GET":
            return None

        return handle_single_item_request(user_context, item, permissions, "read")



import os
import requests

from .exceptions import NonUniqueException, NotFoundException


class Client:
    def __init__(
        self, elody_collection_url=None, static_jwt=None, extra_headers=None, proxy=None
    ):
        self.elody_collection_url = elody_collection_url or os.environ.get(
            "ELODY_COLLECTION_URL", None
        )
        self.static_jwt = static_jwt or os.environ.get("STATIC_JWT", None)
        self.headers = {"Authorization": f"Bearer {self.static_jwt}"}
        if extra_headers:
            self.headers = {**self.headers, **extra_headers}
        self.proxies = None
        if proxy:
            self.proxies = {
                "https": proxy,
                "http": proxy,
            }

    def __create_mediafile(self, entity_id, mediafile):
        url = f"{self.elody_collection_url}/entities/{entity_id}/mediafiles"
        headers = {**self.headers, **{"Accept": "text/uri-list"}}
        response = requests.post(
            url, json=mediafile, headers=headers, proxies=self.proxies
        )
        return self.__handle_response(response, "Failed to create mediafile", "text")

    def create_mediafile_with_filename(
        self,
        filename,
        technical_origin="original",
        original_filename=None,
        institution_id=None,
    ):
        original_filename = original_filename or filename
        data = {
            "filename": filename,
            "original_filename": original_filename,
            "type": "mediafile",
            "technical_origin": technical_origin,
        }
        if institution_id:
            data.update({"metadata": [{"key": "institution", "value": institution_id}]})
        req = requests.post(
            f"{self.elody_collection_url}/mediafiles",
            json=data,
            headers=self.headers,
            proxies=self.proxies,
        )
        if req.status_code != 201:
            raise Exception(req.text.strip())
        return req

    def create_ticket(self, mediafile_name):
        req = requests.post(
            f"{self.elody_collection_url}/tickets",
            json={"filename": mediafile_name},
            headers=self.headers,
            proxies=self.proxies,
        )
        if req.status_code != 201:
            raise Exception(req.text.strip())
        return req.text.strip().replace('"', "")

    def __get_upload_location(
        self,
        entity_id,
        filename,
        is_public=True,
        identifiers=None,
        mediafile_object=None,
    ):
        if not identifiers:
            identifiers = list()
        if not mediafile_object:
            mediafile_object = dict()
        metadata = []
        if is_public:
            metadata = [
                {
                    "key": "publication_status",
                    "value": "publiek",
                }
            ]
        mediafile = {
            **{
                "filename": filename,
                "metadata": metadata,
                "identifiers": identifiers,
            },
            **mediafile_object,
        }
        return self.__create_mediafile(entity_id, mediafile)

    def __handle_response(self, response, error_message, response_type="json"):
        if response.status_code == 409:
            raise NonUniqueException(response.text.strip())
        if response.status_code == 404:
            raise NotFoundException(response.text.strip())
        if response.status_code not in range(200, 300):
            raise Exception(f"{error_message}: {response.text.strip()}")
        match response_type:
            case "json":
                return response.json()
            case "text":
                return response.text.strip()
            case _:
                return response.json()

    def add_entity_mediafiles(self, identifier, payload):
        url = f"{self.elody_collection_url}/entities/{identifier}/mediafiles"
        response = requests.post(
            url, json=payload, headers=self.headers, proxies=self.proxies
        )
        return self.__handle_response(response, "Failed to add mediafiles")

    def add_object(self, collection, payload, params=None):
        url = f"{self.elody_collection_url}/{collection}"
        response = requests.post(
            url, json=payload, headers=self.headers, params=params, proxies=self.proxies
        )
        return self.__handle_response(response, "Failed to add object")

    def add_object_metadata(self, collection, identifier, payload):
        if collection == "entities":
            url = f"{self.elody_collection_url}/{collection}/{identifier}/metadata"
            payload = payload if isinstance(payload, list) else [payload]
            response = requests.patch(
                url, json=payload, headers=self.headers, proxies=self.proxies
            )
            if response.status_code == 400 and response.json()["message"].endswith(
                "has no metadata"
            ):
                response = requests.post(
                    url, json=payload, headers=self.headers, proxies=self.proxies
                )
            return self.__handle_response(response, "Failed to add metadata")
        else:
            url = f"{self.elody_collection_url}/{collection}/{identifier}"
            payload = {"metadata": payload if isinstance(payload, list) else [payload]}
            response = requests.patch(
                url, json=payload, headers=self.headers, proxies=self.proxies
            )
            return self.__handle_response(response, "Failed to add metadata")

    def delete_object(self, collection, identifier):
        url = f"{self.elody_collection_url}/{collection}/{identifier}"
        response = requests.delete(url, headers=self.headers, proxies=self.proxies)
        return self.__handle_response(response, "Failed to delete object", "text")

    def get_all_objects(self, collection):
        url = f"{self.elody_collection_url}/{collection}"
        response = requests.get(url, headers=self.headers, proxies=self.proxies)
        return self.__handle_response(response, "Failed to get objects")

    def get_mediafiles_and_check_existence(self, mediafile_ids):
        mediafile_image_data = []
        for mediafile_id in mediafile_ids:
            response = self.get_object("mediafiles", mediafile_id)
            mediafile_image_data.append(response)
        return mediafile_image_data

    def get_object(self, collection, identifier):
        url = f"{self.elody_collection_url}/{collection}/{identifier}"
        response = requests.get(url, headers=self.headers, proxies=self.proxies)
        return self.__handle_response(response, "Failed to get object")

    def update_object(self, collection, identifier, payload, overwrite=True):
        url = f"{self.elody_collection_url}/{collection}/{identifier}"
        if overwrite:
            response = requests.put(
                url, json=payload, headers=self.headers, proxies=self.proxies
            )
        else:
            response = requests.patch(
                url, json=payload, headers=self.headers, proxies=self.proxies
            )
        return self.__handle_response(response, "Failed to update object")

    def update_object_relations(self, collection, identifier, payload):
        url = f"{self.elody_collection_url}/{collection}/{identifier}/relations"
        response = requests.patch(
            url, json=payload, headers=self.headers, proxies=self.proxies
        )
        return self.__handle_response(response, "Failed to update object relations")

    def upload_file_from_url(
        self,
        entity_id,
        filename,
        file_url,
        identifiers=None,
        upload_location_replace_map=None,
        mediafile_object=None,
    ):
        if not identifiers:
            identifiers = list()
        if not upload_location_replace_map:
            upload_location_replace_map = dict()
        upload_location = self.__get_upload_location(
            entity_id, filename, True, identifiers, mediafile_object
        )
        for current_location, new_location in upload_location_replace_map.items():
            upload_location = upload_location.replace(current_location, new_location)
        print(upload_location)
        mediafile = requests.get(file_url, proxies=self.proxies).content
        response = requests.post(
            upload_location,
            files={"file": mediafile},
            headers=self.headers,
            proxies=self.proxies,
        )
        return self.__handle_response(response, "Failed to upload mediafile")



from configuration import get_object_configuration_mapper  # pyright: ignore
from elody.error_codes import ErrorCode, get_error_code, get_read
from serialization.serialize import serialize  # pyright: ignore
from werkzeug.exceptions import NotFound


def get_content(item, request, content):
    return serialize(
        content,
        type=item.get("type"),
        from_format=serialize.get_format(
            (request.view_args or {}).get("spec", "elody"), request.args
        ),
        to_format=get_object_configuration_mapper().get(item["type"]).SCHEMA_TYPE,
    )


def get_item(storage_manager, user_context_bag, view_args) -> dict:
    view_args = view_args or {}
    if id := view_args.get("id"):
        resolve_collections = user_context_bag.get("collection_resolver")
        collections = resolve_collections(collection=view_args.get("collection"), id=id)
        for collection in collections:
            if item := storage_manager.get_db_engine().get_item_from_collection_by_id(
                collection, id
            ):
                return item

    raise NotFound(
        f"{get_error_code(ErrorCode.ITEM_NOT_FOUND, get_read())} | id:{id} - Item with id {id} does not exist."
    )



from abc import ABC, abstractmethod
from copy import deepcopy
from elody.migration.base_object_migrator import BaseObjectMigrator


class BaseObjectConfiguration(ABC):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    @abstractmethod
    def crud(self):
        return {
            "creator": lambda post_body, **kwargs: post_body,
            "document_content_patcher": lambda *, document, content, overwrite=False, **kwargs: self._document_content_patcher(
                document=document,
                content=content,
                overwrite=overwrite,
                **kwargs,
            ),
            "nested_matcher_builder": lambda object_lists, keys_info, value, **kwargs: self.__build_nested_matcher(
                object_lists, keys_info, value, **kwargs
            ),
            "post_crud_hook": lambda **kwargs: None,
            "pre_crud_hook": lambda *, document, **kwargs: document,
            "storage_type": "db",
        }

    @abstractmethod
    def document_info(self):
        return {}

    @abstractmethod
    def logging(self, flat_document, **kwargs):
        info_labels = {
            "uuid": flat_document.get("_id"),
            "type": flat_document.get("type"),
            "schema": f"{flat_document.get('schema.type')}:{flat_document.get('schema.version')}",
        }
        try:
            from policy_factory import get_user_context  # pyright: ignore

            user_context = get_user_context()
            info_labels["http_method"] = user_context.bag.get("http_method")
            info_labels["requested_endpoint"] = user_context.bag.get(
                "requested_endpoint"
            )
            info_labels["full_path"] = user_context.bag.get("full_path")
            info_labels["preferred_username"] = user_context.preferred_username
            info_labels["email"] = user_context.email
            info_labels["user_roles"] = ", ".join(user_context.x_tenant.roles)
            info_labels["x_tenant"] = user_context.x_tenant.id
        except Exception:
            pass
        return {"info_labels": info_labels, "loki_indexed_info_labels": {}}

    @abstractmethod
    def migration(self):
        return BaseObjectMigrator(status="disabled")

    @abstractmethod
    def serialization(self, from_format, to_format):
        def serializer(document, **_):
            return document

        return serializer

    @abstractmethod
    def validation(self):
        def validator(http_method, content, item, **_):
            pass

        return "function", validator

    def _document_content_patcher(
        self, *, document, content, overwrite=False, **kwargs
    ):
        raise NotImplementedError(
            "Provide concrete implementation in child object configuration"
        )

    def _merge_object_lists(self, source, target, object_list_key):
        for target_item in target:
            for source_item in source:
                if source_item[object_list_key] == target_item[object_list_key]:
                    source.remove(source_item)
        return [*source, *target]

    def _get_user_context_id(self):
        try:
            from policy_factory import get_user_context  # pyright: ignore

            return get_user_context().id
        except Exception:
            return None

    def _sanitize_document(self, *, document, **kwargs):
        sanitized_document = {}
        document_deepcopy = deepcopy(document)
        for key, value in document_deepcopy.items():
            if isinstance(value, dict):
                sanitized_value = BaseObjectConfiguration._sanitize_document(
                    self, document=value
                )
                if sanitized_value:
                    sanitized_document[key] = sanitized_value
            elif isinstance(value, list):
                sanitized_document[key] = [
                    element.strip() if isinstance(element, str) else element
                    for element in value
                    if element
                ]
                if all(isinstance(element, str) for element in sanitized_document[key]):
                    sanitized_document[key] = list(set(sanitized_document[key]))
            elif isinstance(value, str):
                lines = value.splitlines()
                value = "\n".join(line.strip() for line in lines).strip()
                sanitized_document[key] = value.strip()
            elif value:
                sanitized_document[key] = value
        return sanitized_document

    def _should_create_history_object(self):
        try:
            from policy_factory import get_user_context  # pyright: ignore

            get_user_context()
            return bool(self.crud().get("collection_history"))
        except Exception:
            return False

    def _sort_document_keys(self, document):
        def sort_keys(data):
            if isinstance(data, dict):
                sorted_items = {key: data.pop(key) for key in sorted(data.keys())}
                for key, value in sorted_items.items():
                    data[key] = sort_keys(value)
                return data
            elif isinstance(data, list):
                if all(isinstance(i, str) for i in data):
                    data.sort()
                    return data
                else:
                    for index, item in enumerate(data):
                        data[index] = sort_keys(item)
                    return data
            else:
                return data

        for key, value in self.document_info().get("object_lists", {}).items():
            if document.get(key):
                document[key] = sorted(
                    document[key], key=lambda property: property[value]
                )
        sort_keys(document)
        return document

    def __build_nested_matcher(
        self, object_lists, keys_info, value, *, index=0, **kwargs
    ):
        if index == 0 and not any(info["object_list"] for info in keys_info):
            if value in ["ANY_MATCH", "NONE_MATCH"]:
                value = {"$exists": value == "ANY_MATCH"}
            matcher = {".".join(info["key"] for info in keys_info): value}
            if inner_exact_matches := kwargs.get("inner_exact_matches"):
                matcher.update(inner_exact_matches)
            return matcher

        info = keys_info[index]

        if info["object_list"]:
            elem_match = {
                "$elemMatch": {
                    object_lists[info["object_list"]]: info["object_key"],
                    **self.__build_nested_matcher(
                        object_lists, keys_info[index + 1 :], value, index=0, **kwargs
                    ),
                }
            }
            if value in ["ANY_MATCH", "NONE_MATCH"]:
                elem_match_with_exists_operator = deepcopy(elem_match)
                del elem_match["$elemMatch"][keys_info[index + 1]["key"]]
                if value == "NONE_MATCH":
                    return {
                        "NOR_MATCHER": [
                            {info["key"]: {"$all": [elem_match]}},
                            {info["key"]: {"$all": [elem_match_with_exists_operator]}},
                        ]
                    }
            return elem_match if index > 0 else {info["key"]: {"$all": [elem_match]}}

        raise Exception(f"Unable to build nested matcher. See keys_info: {keys_info}")
